@@ -25,6 +25,7 @@ static struct servants_list_item *servants_leader = NULL;
 int     disk_priority = 1;
 int	check_pcmk = 1;
 int	check_cluster = 1;
+int	has_check_pcmk_env = false;
 int	disk_count	= 0;
 int	servant_count	= 0;
 int	servant_restart_interval = 5;
@@ -39,6 +40,57 @@ bool sync_resource_startup = false;
 
 int parse_device_line(const char *line);
 
+static int
+sanitize_numeric_option_value(const char *value)
+{
+    char *end = NULL;
+    long int result = -1;
+
+    if (value == NULL) {
+        return -1;
+    }
+
+    errno = 0;
+
+    result = strtol(value, &end, 10);
+    if (result <= INT_MIN || result >= INT_MAX || errno != 0) {
+        result = -1;
+    } else if (*end != '\0') {
+        result = -1;
+    }
+
+    return (int)result;
+}
+
+static const char *
+sanitize_option_value(const char *value)
+{
+	size_t max = 0;
+	size_t lpc = 0;
+
+	if (value == NULL) {
+		return NULL;
+	}
+
+	max = strlen(value);
+
+	for (lpc = 0; lpc < max; lpc++) {
+		if (!isspace(value[lpc])) {
+			break;
+		}
+	}
+
+	return (strlen(value + lpc) > 0 ? (value + lpc) : NULL);
+}
+
+static const char *
+get_env_option(const char *option)
+{
+	const char *value = getenv(option);
+
+	return sanitize_option_value(value);
+}
+
 static int
 recruit_servant(const char *devname, pid_t pid)
 {
@@ -698,7 +750,7 @@ void inquisitor_child(void)
 
 		if (timeout_watchdog_warn && (latency > (int)timeout_watchdog_warn)) {
 			cl_log(LOG_WARNING,
-			       "Latency: No liveness for %d s exceeds threshold of %d s (healthy servants: %d)",
+			       "Latency: No liveness for %ds exceeds watchdog warning timeout of %ds (healthy servants: %d)",
 			       (int)latency, (int)timeout_watchdog_warn, good_servants);
 
                         if (debug_mode && watchdog_use) {
@@ -894,14 +946,16 @@ int main(int argc, char **argv, char **envp)
 
 	sbd_get_uname();
 
-        value = getenv("SBD_PACEMAKER");
+        value = get_env_option("SBD_PACEMAKER");
         if(value) {
             check_pcmk = crm_is_true(value);
             check_cluster = crm_is_true(value);
-        }
-        cl_log(LOG_INFO, "Enable pacemaker checks: %d (%s)", (int)check_pcmk, value?value:"default");
 
-        value = getenv("SBD_STARTMODE");
+            has_check_pcmk_env = true;
+        }
+        cl_log(LOG_INFO, "SBD_PACEMAKER set to: %d (%s)", (int)check_pcmk, value?value:"default");
+
+        value = get_env_option("SBD_STARTMODE");
         if(value == NULL) {
         } else if(strcmp(value, "clean") == 0) {
             start_mode = 1;
@@ -910,7 +964,7 @@ int main(int argc, char **argv, char **envp)
         }
         cl_log(LOG_INFO, "Start mode set to: %d (%s)", (int)start_mode, value?value:"default");
 
-        value = getenv("SBD_WATCHDOG_DEV");
+        value = get_env_option("SBD_WATCHDOG_DEV");
         if(value) {
             free(watchdogdev);
             watchdogdev = strdup(value);
@@ -919,26 +973,23 @@ int main(int argc, char **argv, char **envp)
 
         /* SBD_WATCHDOG has been dropped from sbd.sysconfig example.
          * This is for backward compatibility. */
-        value = getenv("SBD_WATCHDOG");
+        value = get_env_option("SBD_WATCHDOG");
         if(value) {
             watchdog_use = crm_is_true(value);
         }
 
-        value = getenv("SBD_WATCHDOG_TIMEOUT");
+        value = get_env_option("SBD_WATCHDOG_TIMEOUT");
         if(value) {
             timeout_watchdog = crm_get_msec(value) / 1000;
-            if(timeout_watchdog > 5) {
-                timeout_watchdog_warn = (int)timeout_watchdog / 5 * 3;
-            }
         }
 
-        value = getenv("SBD_PIDFILE");
+        value = get_env_option("SBD_PIDFILE");
         if(value) {
             pidfile = strdup(value);
             cl_log(LOG_INFO, "pidfile set to %s", pidfile);
         }
 
-        value = getenv("SBD_DELAY_START");
+        value = get_env_option("SBD_DELAY_START");
         if(value) {
             delay_start = crm_is_true(value);
 
@@ -949,17 +1000,13 @@ int main(int argc, char **argv, char **envp)
                 }
             }
         }
-        cl_log(LOG_DEBUG, "Delay start: %s%s%s",
-               delay_start? "yes (" : "no",
-               delay_start? (delay > 0 ? value: "msgwait") : "",
-               delay_start? ")" : "");
 
-        value = getenv("SBD_TIMEOUT_ACTION");
+        value = get_env_option("SBD_TIMEOUT_ACTION");
         if(value) {
             timeout_action = strdup(value);
         }
 
-        value = getenv("SBD_MOVE_TO_ROOT_CGROUP");
+        value = get_env_option("SBD_MOVE_TO_ROOT_CGROUP");
         if(value) {
             move_to_root_cgroup = crm_is_true(value);
 
@@ -972,26 +1019,19 @@ int main(int argc, char **argv, char **envp)
             }
         }
 
-        value = getenv("SBD_SYNC_RESOURCE_STARTUP");
-        sync_resource_startup =
-            crm_is_true(value?value:SBD_SYNC_RESOURCE_STARTUP_DEFAULT);
-
-#if !USE_PACEMAKERD_API
-        if (sync_resource_startup) {
-            fprintf(stderr, "Failed to sync resource-startup as "
-                "SBD was built against pacemaker not supporting pacemakerd-API.\n");
-            exit_status = -1;
-            goto out;
-        }
-#else
-        if (!sync_resource_startup) {
-            cl_log(LOG_WARNING, "SBD built against pacemaker supporting "
-                             "pacemakerd-API. Should think about enabling "
-                             "SBD_SYNC_RESOURCE_STARTUP.");
-        }
-#endif
-
 	while ((c = getopt(argc, argv, "czC:DPRTWZhvw:d:n:p:1:2:3:4:5:t:I:F:S:s:r:")) != -1) {
+		int sanitized_num_optarg = 0;
+		/* Call it before checking optarg for NULL to make coverity happy */
+		const char *sanitized_optarg = sanitize_option_value(optarg);
+
+		if (optarg && ((sanitized_optarg == NULL) ||
+				(strchr("SsC12345tIF", c) &&
+				(sanitized_num_optarg = sanitize_numeric_option_value(sanitized_optarg)) < 0))) {
+			fprintf(stderr, "Invalid value \"%s\" for option -%c\n", optarg, c);
+			exit_status = -2;
+			goto out;
+		}
+
 		switch (c) {
 		case 'D':
 			break;
@@ -1004,11 +1044,11 @@ int main(int argc, char **argv, char **envp)
 			cl_log(LOG_INFO, "Realtime mode deactivated.");
 			break;
 		case 'S':
-			start_mode = atoi(optarg);
+			start_mode = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Start mode set to: %d", (int)start_mode);
 			break;
 		case 's':
-			timeout_startup = atoi(optarg);
+			timeout_startup = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Start timeout set to: %d", (int)timeout_startup);
 			break;
 		case 'v':
@@ -1035,14 +1075,14 @@ int main(int argc, char **argv, char **envp)
 			W_count++;
 			break;
 		case 'w':
-                        cl_log(LOG_NOTICE, "Using watchdog device '%s'", watchdogdev);
                         free(watchdogdev);
-                        watchdogdev = strdup(optarg);
+                        watchdogdev = strdup(sanitized_optarg);
                         watchdogdev_is_default = false;
+                        cl_log(LOG_NOTICE, "Using watchdog device '%s'", watchdogdev);
 			break;
 		case 'd':
 #if SUPPORT_SHARED_DISK
-			if (recruit_servant(optarg, 0) != 0) {
+			if (recruit_servant(sanitized_optarg, 0) != 0) {
 				fprintf(stderr, "Invalid device: %s\n", optarg);
 				exit_status = -1;
 				goto out;
@@ -1063,50 +1103,48 @@ int main(int argc, char **argv, char **envp)
 			disk_priority = 0;
 			break;
 		case 'n':
-			local_uname = strdup(optarg);
+			local_uname = strdup(sanitized_optarg);
 			cl_log(LOG_INFO, "Overriding local hostname to %s", local_uname);
 			break;
 		case 'p':
-			pidfile = strdup(optarg);
+			pidfile = strdup(sanitized_optarg);
 			cl_log(LOG_INFO, "pidfile set to %s", pidfile);
 			break;
 		case 'C':
-			timeout_watchdog_crashdump = atoi(optarg);
+			timeout_watchdog_crashdump = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Setting crashdump watchdog timeout to %d",
 					(int)timeout_watchdog_crashdump);
 			break;
 		case '1':
-			timeout_watchdog = atoi(optarg);
-                        if(timeout_watchdog > 5) {
-                            timeout_watchdog_warn = (int)timeout_watchdog / 5 * 3;
-                        }
+			timeout_watchdog = sanitized_num_optarg;
 			break;
 		case '2':
-			timeout_allocate = atoi(optarg);
+			timeout_allocate = sanitized_num_optarg;
 			break;
 		case '3':
-			timeout_loop = atoi(optarg);
+			timeout_loop = sanitized_num_optarg;
 			break;
 		case '4':
-			timeout_msgwait = atoi(optarg);
+			timeout_msgwait = sanitized_num_optarg;
 			break;
 		case '5':
-			timeout_watchdog_warn = atoi(optarg);
+			timeout_watchdog_warn = sanitized_num_optarg;
+			do_calculate_timeout_watchdog_warn = false;
 			cl_log(LOG_INFO, "Setting latency warning to %d",
 					(int)timeout_watchdog_warn);
 			break;
 		case 't':
-			servant_restart_interval = atoi(optarg);
+			servant_restart_interval = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Setting servant restart interval to %d",
 					(int)servant_restart_interval);
 			break;
 		case 'I':
-			timeout_io = atoi(optarg);
+			timeout_io = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Setting IO timeout to %d",
 					(int)timeout_io);
 			break;
 		case 'F':
-			servant_restart_count = atoi(optarg);
+			servant_restart_count = sanitized_num_optarg;
 			cl_log(LOG_INFO, "Servant restart count set to %d",
 					(int)servant_restart_count);
 			break;
@@ -1114,7 +1152,7 @@ int main(int argc, char **argv, char **envp)
 			if (timeout_action) {
 				free(timeout_action);
 			}
-			timeout_action = strdup(optarg);
+			timeout_action = strdup(sanitized_optarg);
 			break;
 		case 'h':
 			usage();
@@ -1132,7 +1170,7 @@ int main(int argc, char **argv, char **envp)
            then it is probably undesirable to add those
            from environment (general rule cmdline has precedence)
          */
-        value = getenv("SBD_DEVICE");
+        value = get_env_option("SBD_DEVICE");
         if ((value) && strlen(value)) {
 #if SUPPORT_SHARED_DISK
             int devices = parse_device_line(value);
@@ -1167,7 +1205,17 @@ int main(int argc, char **argv, char **envp)
 	}
 
 	if (P_count > 0) {
-		check_pcmk = arg_enabled(P_count);
+		int check_pcmk_arg = arg_enabled(P_count);
+
+		if (has_check_pcmk_env && check_pcmk_arg != check_pcmk) {
+			cl_log(LOG_WARNING, "Pacemaker integration is %s: "
+					"SBD_PACEMAKER=%s is overridden by %s option. "
+					"It's recommended to only use SBD_PACEMAKER.",
+					check_pcmk_arg? "enabled" : "disabled",
+					check_pcmk? "yes" : "no",
+					check_pcmk_arg? "-P" : "-PP");
+		}
+		check_pcmk = check_pcmk_arg;
 	}
 
 	if ((disk_count > 0) && (strlen(local_uname) > SECTOR_NAME_MAX)) {
@@ -1228,6 +1276,33 @@ int main(int argc, char **argv, char **envp)
 		}
 	}
 
+    if (strcmp(argv[optind], "watch") == 0) {
+        value = get_env_option("SBD_SYNC_RESOURCE_STARTUP");
+        sync_resource_startup =
+            crm_is_true(value?value:SBD_SYNC_RESOURCE_STARTUP_DEFAULT);
+
+#if !USE_PACEMAKERD_API
+        if (sync_resource_startup) {
+            fprintf(stderr, "Failed to sync resource-startup as "
+                "SBD was built against pacemaker not supporting pacemakerd-API.\n");
+            exit_status = -1;
+            goto out;
+        }
+#else
+        if (check_pcmk && !sync_resource_startup) {
+            cl_log(LOG_WARNING, "SBD built against pacemaker supporting "
+                             "pacemakerd-API. Should think about enabling "
+                             "SBD_SYNC_RESOURCE_STARTUP.");
+
+        } else if (!check_pcmk && sync_resource_startup) {
+            fprintf(stderr, "Set SBD_PACEMAKER=yes to allow resource startup syncing. "
+                    "Otherwise explicitly set SBD_SYNC_RESOURCE_STARTUP=no if to intentionally disable.\n");
+            exit_status = -1;
+            goto out;
+        }
+#endif
+    }
+
 #if SUPPORT_SHARED_DISK
 	if (strcmp(argv[optind], "create") == 0) {
 		exit_status = init_devices(servants_leader);
@@ -1247,25 +1322,8 @@ int main(int argc, char **argv, char **envp)
         } else if (strcmp(argv[optind], "ping") == 0) {
             exit_status = ping_via_slots(argv[optind + 1], servants_leader);
 
-        } else if (strcmp(argv[optind], "watch") == 0) {
-                if(disk_count > 0) {
-                    /* If no devices are specified, its not an error to be unable to find one */
-                    open_any_device(servants_leader);
-                }
-
-                if (delay_start) {
-                    if (delay <= 0) {
-                        delay = get_first_msgwait(servants_leader);
-                    }
-
-                    sleep((unsigned long) delay);
-                }
-
-	} else {
-		exit_status = -2;
-	}
+        } else
 #endif
-
         if (strcmp(argv[optind], "query-watchdog") == 0) {
             exit_status = watchdog_info();
         } else if (strcmp(argv[optind], "test-watchdog") == 0) {
@@ -1273,6 +1331,45 @@ int main(int argc, char **argv, char **envp)
         } else if (strcmp(argv[optind], "watch") == 0) {
             /* sleep $(sbd $SBD_DEVICE_ARGS dump | grep -m 1 msgwait | awk '{print $4}') 2>/dev/null */
 
+                const char *delay_source = delay ? "SBD_DELAY_START" : "";
+
+#if SUPPORT_SHARED_DISK
+                if(disk_count > 0) {
+                    /* If no devices are specified, its not an error to be unable to find one */
+                    open_any_device(servants_leader);
+
+                    if (delay_start && delay <= 0) {
+                        delay = get_first_msgwait(servants_leader);
+
+                        if (delay > 0) {
+                            delay_source = "msgwait";
+                        } else {
+                            cl_log(LOG_WARNING, "No 'msgwait' value from disk, using '2 * watchdog-timeout' for 'delay' starting");
+                        }
+                    }
+                }
+#endif
+                /* Re-calculate timeout_watchdog_warn based on any timeout_watchdog from:
+                 * SBD_WATCHDOG_TIMEOUT, -1 option or on-disk setting read with open_any_device() */
+                if (do_calculate_timeout_watchdog_warn) {
+                    timeout_watchdog_warn = calculate_timeout_watchdog_warn(timeout_watchdog);
+                }
+
+                if (delay_start) {
+                    /* diskless mode or disk read issues causing get_first_msgwait() to return a 0 for delay */
+                    if (delay <= 0) {
+                        delay = 2 * timeout_watchdog;
+                        delay_source = "watchdog-timeout * 2";
+                    }
+
+                    cl_log(LOG_DEBUG, "Delay start (yes), (delay: %ld), (delay source: %s)", delay, delay_source);
+
+                    sleep((unsigned long) delay);
+
+                } else {
+                    cl_log(LOG_DEBUG, "Delay start (no)");
+                }
+
                 /* We only want this to have an effect during watch right now;
                  * pinging and fencing would be too confused */
                 cl_log(LOG_INFO, "Turning on pacemaker checks: %d", check_pcmk);
@@ -1291,6 +1388,8 @@ int main(int argc, char **argv, char **envp)
                 cl_log(LOG_NOTICE, "%s flush + write \'%c\' to sysrq in case of timeout",
                        do_flush?"Do":"Skip", timeout_sysrq_char);
                 exit_status = inquisitor();
+        } else {
+            exit_status = -2;
         }
         
   out:
